
#version 120

uniform vec4 sl_fogColorAndDensity;
uniform vec4 sl_cloudTint;
uniform sampler2D sl_tex2D;
uniform sampler2D sl_tex2D2;

uniform vec4 sl_displacementVectorAndContrast;
uniform vec4 sl_fadeAndDisplacementFactor;
uniform vec4 sl_upVectorAndThickness;
uniform vec4 sl_sunColor;
uniform vec4 sl_skyColor;
uniform vec4 sl_lightDirection;
uniform vec4 sl_layerThicknessAndIsTop;
uniform vec4 sl_layerSizeAndUnitScale;
uniform mat4 sl_invBasis;
uniform vec4 sl_groundColor;
uniform vec4 sl_extinctionFactor;
uniform vec4 sl_outputScale;

varying vec3 V;
varying float fadeDist;

void overrideStratusLighting(in vec3 fogColor, in float fogFactor, in vec3 cloudColor, in float cloudFade, inout vec4 finalColor);
void writeFragmentData(in vec4 finalColor);

#define PI 3.14159265
#define extinctionCoefficient 0.04618
#define EPSILON_ABOVE 0.1
#define EPSILON_BELOW 0.05
#define EDGE_TRANSLUCENCY_ABOVE 0.3
#define EDGE_TRANSLUCENCY_BELOW 0.15
#define MIN_OCCLUSION 0.2

// Artificial boost to make fogbow & glory more visible
#define SINGLE_SCATTERING_SCALE 1.7

// Sacrifice some accuracy for speed.
//#define FAST

#define MIE

float getH(vec4 texel)
{
    return  texel.y * sl_layerThicknessAndIsTop.x;
}

vec3 phaseFunctionHG(float costheta, float g)
{
    float hg = (1.0 - g * g) / pow(1.0 + g * g - 2.0 * g * costheta, 1.5);
    return vec3(hg,hg,hg);
}

vec3 phaseFunctionMie(float theta)
{
    float u = theta / PI;   // 0-1 => 0 - 180
    u = u * (1801.0 / 2048.0);
    return texture2D(sl_tex2D2, vec2(u, 0.0)).xyz;
}

vec3 phaseFunctionMieConvolved(float theta)
{
    float u = theta / PI;   // 0-1 => 0 - 180
    u = u * (1801.0 / 2048.0);
    vec3 p = texture2D(sl_tex2D2, vec2(u, 1.0)).xyz;
    return p;
}

float computeTransparency(float x, float fudge)
{
    //const float effectiveRadius = 0.000007; //  AKA re
    //const float densityOfDroplets = 300000000.0; // AKA N0

    //float extinctionCrossSection = PI * effectiveRadius * effectiveRadius; // AKA sigma
    //float extinctionCoefficient = densityOfDroplets * extinctionCrossSection; // AKA kappa
    float extinction = exp(-extinctionCoefficient * x * sl_extinctionFactor.x * fudge);

    return extinction;
}

void computeSingleScatteringReflectance(in float Hv, in float Hl, in float ul, in float uv, in float vl, out vec3 Ir1, out vec3 It1)
{
#ifdef MIE
    vec3 phase = phaseFunctionMie(acos(vl));
#else
    vec3 phase = phaseFunctionHG(vl, 0.99);
#endif

    vec3 t1 = (extinctionCoefficient * phase * ul) / (uv + ul);
    float t2 = 1.0 - computeTransparency(Hl + Hv, 1.0);

    Ir1 = t1 * (t2 * SINGLE_SCATTERING_SCALE);

    t1 = (extinctionCoefficient * phase * ul) / (uv - ul);
    t2 = computeTransparency(Hv, 1.0) - computeTransparency(Hl, 1.0);

    It1 = t1 * (t2 * SINGLE_SCATTERING_SCALE);
}

void computeDoubleScatteringReflectance(in float Hv, in float Hl, in float ul, in float uv, in float vl, out vec3 Ir2, out vec3 It2)
{
#ifdef MIE
    vec3 phase = phaseFunctionMieConvolved(acos(vl));
#else
    vec3 phase = phaseFunctionHG(vl, 0.99 * 0.99);
#endif
    vec3 t1 = (extinctionCoefficient * extinctionCoefficient  * phase * ul) / (uv + ul);
    float t2 = 1.0 - computeTransparency(Hl + Hv, 1.0);

    Ir2 = t1 * t2;

    t1 = (extinctionCoefficient * extinctionCoefficient * phase * ul) / (uv - ul);
    t2 = computeTransparency(Hv, 1.0) - computeTransparency(Hl, 1.0);

    It2 = t1 * t2;
}

void computeMultipleScatteringReflectance(in float H, in float Hl, in float ul, out float Rms, out float Tms, out float R3plus, out float T3plus)
{
#ifdef FAST
    float b = 1.0;
    float c = 0.02;
    float kc = 0.03;
    float t = 0.7;
    float r = 0.06;
#else
    float l = acos(ul);
    float u = l / PI;   // 0-1 => 0 - 180
    u = u * (10.0 / 2048.0);
    u += (2000.0 / 2048.0);
    vec3 k1 = texture2D(sl_tex2D2, vec2(u, 0.0)).xyz;
    vec2 k2 = texture2D(sl_tex2D2, vec2(u, 1.0)).xy;

    float b = k1.x;
    float c = k1.y;
    float kc = k1.z;
    float t = k2.x;
    float r = k2.y;
#endif

    Tms = 0.9961 / (H - (H - 1.0) * 0.9961);

    Tms *= b + (1.0 - b) * exp(-c * H);

    Rms = 1.0 - Tms;

    float T0 = exp(-kc * Hl);

    float R1 = 0.5 * r * (1.0 - exp(-kc * 2.0 * H));
    float T1 = t * H * kc * exp(-kc * H);

    float R2 = 0.5 * r * t * (1.0 - (2.0 * H * kc + 1.0) * exp(-kc * 2.0 * H));

    float T2a = 0.5 * pow(kc * t * H, 2.0) * exp(-kc * H);
    float T2b = 0.25 * r * r * (1.0 + (2.0 * H * kc - 1.0) * exp(-kc * 2.0 * H)) * exp(-kc * 3.0 * H);
    float T2 = T2a + T2b;

    R3plus = Rms - R1 - R2;
    T3plus = Tms - T0 - T1 - T2;
}

void toneMap(inout vec3 c)
{
    float A = 0.9;
    float gamma = 0.65;
    c = vec3(A, A, A) * pow(c, vec3(gamma, gamma, gamma));
}

void main()
{
    float unitScale = sl_layerSizeAndUnitScale.z;

    float vertDist = length(V) * unitScale;

    float fadeFalloff = sl_layerSizeAndUnitScale.w;
    float cloudFade = 1.0;

    if (fadeFalloff > 0) {
        float len = fadeDist / (sl_layerSizeAndUnitScale.x * 0.5);
        if (len > 1.0) {
            discard;
            return;
        }
        cloudFade = exp(-fadeFalloff * len * len * len) * sl_fadeAndDisplacementFactor.x;
    }

    vec4 fogColor = vec4(sl_fogColorAndDensity.xyz, cloudFade);
    float fogExponent = vertDist * sl_fogColorAndDensity.w;
    float fogFactor = clamp(exp(-(fogExponent * fogExponent)), 0.0, 1.0);

    vec4 texel = texture2D(sl_tex2D, gl_TexCoord[0].st);

    float H = getH(texel) * unitScale;

    float coverage = sl_fadeAndDisplacementFactor.z;
    float ambientOcclusion = 1.0;
    if (coverage >= 1.0) {
        float scudThickness = sl_layerThicknessAndIsTop.y;
        float layerThickness = sl_layerThicknessAndIsTop.x;
        float minNSrc = sl_extinctionFactor.y;
        float thicknessFactor = texel.x * (1.0 - minNSrc) + minNSrc;
        float scudDepth = (layerThickness - (thicknessFactor * layerThickness)) / scudThickness;
        scudDepth = max(EPSILON_ABOVE, scudDepth);
        ambientOcclusion = atan(1.0 / scudDepth) / (PI * 0.5);
        ambientOcclusion = max(MIN_OCCLUSION, ambientOcclusion);
    }

    vec3 cloudColor;
    vec4 finalColor;

    vec3 Lnorm = normalize((sl_invBasis * sl_lightDirection).xyz);
    vec3 Vnorm = normalize(V);

    vec3 normal = normalize(sl_upVectorAndThickness.xyz);
    bool above = dot(Vnorm, normal) >= 0;

    // Due to curvature of the cloud, the dot product isn't a good test when below.
    if (sl_layerThicknessAndIsTop.w < 1.0) {
        above = sl_layerThicknessAndIsTop.z > 0;
    }

    //bool inside = (sl_layerThicknessAndIsTop.w > 0 && !above) || (sl_layerThicknessAndIsTop.w < 1.0 && above);

    float uv, ul, vl, uvc, ulc;
    float Hv, Hl;
    float edgeTranslucency;

    if (!above) {
        uv = dot(Vnorm, -normal);
        ul = dot(Lnorm, normal);
        uvc = max(EPSILON_BELOW, uv);
        ulc = max(EPSILON_BELOW, ul);
        vl = dot(-Vnorm, Lnorm);
        Hv = H / uvc;
        Hl = H / ulc;
        edgeTranslucency = EDGE_TRANSLUCENCY_BELOW;
    } else {
        float NORMAL_OFF = (1.0 / 2048.0);
        vec3 off = vec3(-NORMAL_OFF, 0, NORMAL_OFF);

        vec2 texturePos = gl_TexCoord[0].st;

        // s11 = Current
        float s11 = H;

        // s21 = Right
        float s21 = getH(texture2D(sl_tex2D, vec2(texturePos.xy + off.zy))) * unitScale;

        // s12 = Above
        float s12 = getH(texture2D(sl_tex2D, vec2(texturePos.xy + off.yz))) * unitScale;

        vec3 va = normalize( vec3(off.z * sl_layerSizeAndUnitScale.x, s21 - s11, 0.0) );
        vec3 vb = normalize( vec3(0.0, s12 - s11, off.z * sl_layerSizeAndUnitScale.y) );

        vec3 localNormal = normalize( cross(vb, va) );
        vec4 l4 = sl_invBasis * vec4(localNormal, 1.0);
        localNormal = l4.xyz / l4.w;

        uv = max(EPSILON_ABOVE,dot(Vnorm, localNormal));
        ul = max(EPSILON_ABOVE+0.001f,dot(Lnorm, localNormal));

        vl = dot(-Vnorm, Lnorm);
        Hv = H / uv;
        Hl = H / ul;
        edgeTranslucency = EDGE_TRANSLUCENCY_ABOVE;
    }

//    float Hv = inside ? min(H / uv, vertDist) : H / uv;


    float Ir3plus;
    float It0, It3plus;
    float Rms, Tms, R3plus, T3plus;
    vec3 Ir1, It1, Ir2, It2;

    It0 = computeTransparency(Hv, edgeTranslucency / sl_extinctionFactor.x);

    float a = 1.0 - It0;

    computeSingleScatteringReflectance(Hv, Hl, ul, uv, vl, Ir1, It1);
    computeDoubleScatteringReflectance(Hv, Hl, ul, uv, vl, Ir2, It2);

    if (above) {
        uv = max(EPSILON_ABOVE,dot(Vnorm, normal));
        ul = max(EPSILON_ABOVE,dot(Lnorm, normal));
    } else {
        uv = max(EPSILON_BELOW,dot(Vnorm, -normal));
        ul = max(EPSILON_BELOW,dot(Lnorm, normal));
    }
    // Hv = inside ? min(H / uv, vertDist) : H / uv;
    Hv = H / uv;
    Hl = H / ul;

    computeMultipleScatteringReflectance(H, Hl, ul, Rms, Tms, R3plus, T3plus);

    It3plus = T3plus * (ul / (4.0 * PI * uv));
    Ir3plus = R3plus * (ul / (4.0 * PI * uv));

    vec3 Ir = Ir1 + Ir2 + vec3(Ir3plus, Ir3plus, Ir3plus);
    vec3 It = vec3(It0, It0, It0) + It1 + It2 + vec3(It3plus, It3plus, It3plus);

    if (above) {
        // Top
        cloudColor = Ir * sl_sunColor.xyz + Rms * sl_skyColor.xyz * ambientOcclusion + Tms * sl_groundColor.xyz;
    } else {
        // Bottom
        cloudColor = It * sl_sunColor.xyz + Tms * sl_skyColor.xyz + Rms * sl_groundColor.xyz;
    }

    toneMap(cloudColor);

    cloudColor *= sl_cloudTint.xyz;

    finalColor.xyz = mix(fogColor.xyz, cloudColor, fogFactor);
    finalColor.a = a * cloudFade;

    if (!above) finalColor.a *= fogFactor;

    overrideStratusLighting(fogColor.xyz, fogFactor, cloudColor, cloudFade, finalColor);

    finalColor.xyz *= sl_outputScale.xxx;

    if (finalColor != finalColor) {
        discard;
    } else {
        writeFragmentData(finalColor);
    }

}
